﻿// Fill out your copyright notice in the Description page of Project Settings.


#include "TurnBaseCombatV2.h"
#include "CoreMinimal.h"
#include "AIController.h"
#include "StatusSystem.h"
#include "Blueprint/UserWidget.h"
#include "BehaviorTree/BlackboardComponent.h"
#include "Components/TextBlock.h"
#include "Components/ProgressBar.h"
#include "GameFramework/Character.h"
#include "Kismet/GameplayStatics.h"
#include "COMP250_1_2101327_AI/Player/PlayerCharacter.h"
#include "UObject/UnrealTypePrivate.h"

// Sets default values
ATurnBaseCombatV2::ATurnBaseCombatV2()
{
	if (HUDWidget == nullptr)
	{
		static ConstructorHelpers::FClassFinder<UUserWidget> HUDWidgetClass(TEXT("/Game/Blueprints/Combat_UI/Combat_UI"));
		HUDWidget = HUDWidgetClass.Class;
		static ConstructorHelpers::FClassFinder<UUserWidget> ActionItemWidgetClass(TEXT("/Game/Blueprints/Combat_UI/ActionItem"));
		ActionPlanItemWidget = ActionItemWidgetClass.Class;
		static ConstructorHelpers::FClassFinder<UStatusEffect> StatusEffectThornsClassFinder(TEXT("/Game/Blueprints/StatusEffects/BP_Thorns"));
		ThornsStatusEffect = StatusEffectThornsClassFinder.Class;
		static ConstructorHelpers::FClassFinder<UStatusEffect> StatusEffectDotClassFinder(TEXT("/Game/Blueprints/StatusEffects/BP_DamageOverTime"));
		DOTStatusEffect = StatusEffectDotClassFinder.Class;

		static ConstructorHelpers::FClassFinder<UUserWidget> WinClassFinder(TEXT("/Game/Blueprints/WinLoseStates/Win"));
		WinWidget = WinClassFinder.Class;
		static ConstructorHelpers::FClassFinder<UUserWidget> LoseClassFinder(TEXT("/Game/Blueprints/WinLoseStates/Lose"));
		LoseWidget = LoseClassFinder.Class;
		//static ConstructorHelpers::FClassFinder<UUserWidget> DeathScreenWidgetClass(TEXT("/Game/Blueprints/Death_UI/Death_UI"));
		//DeathScreenWidgetSubclass = DeathScreenWidgetClass.Class;
	}
}

void ATurnBaseCombatV2::StartCombat(AActor* Enemy)
{
	if (Enemy == nullptr) return;
	const FProperty* HealthProperty = Enemy->GetClass()->FindPropertyByName(FName("Health"));
	int32* EnemyHealthPtr = HealthProperty->ContainerPtrToValuePtr<int32>(Enemy);
	EnemyHealth = EnemyHealthPtr;

	if (HUD->IsInViewport()) return;
	HUD->AddToViewport();
	EnemyActor = Enemy;

	ProbertiumResource = FMath::RandRange(0, 10);
	EisResource = FMath::RandRange(0, 10);
	AzosResource = FMath::RandRange(0, 10);
	IroquoidResource = FMath::RandRange(0, 10);

	EnemyProbertiumResource = FMath::RandRange(0, 10);
	EnemyEisResource = FMath::RandRange(0, 10);
	EnemyAzosResource = FMath::RandRange(0, 10);
	EnemyIroquoidResource = FMath::RandRange(0, 10);

	//Disable Character Movement
	if (ACharacter* PlayerCharacter = Cast<ACharacter>(GetWorld()->GetFirstPlayerController()->GetPawn()))
	{
		PlayerCharacter->DisableInput(GetWorld()->GetFirstPlayerController());
	}
	//Set to UI Mode Only
	APlayerController* PlayerController = GetWorld()->GetFirstPlayerController();
	PlayerController->SetInputMode(FInputModeUIOnly());
	PlayerController->bShowMouseCursor = true;

	CurrentComboString = "";
	UpdateComboString(CurrentComboString);
	RevertActionPoints();
	UpdateActionPoints();
	UpdateResourceBars();
	ClearBattleLog();

	EnemyGOAPAgent = Cast<UGOAPAgent>(EnemyActor->GetComponentByClass(UGOAPAgent::StaticClass()));
	EnemyActionPlan = EnemyGOAPAgent->Plan(GetWorldState(), EnemyGOAPAgent->Goals);
	UpdateActionPlanWidget();
}

void ATurnBaseCombatV2::EndCombat()
{
	bEnemyHasExtraTurn = false;
	bPlayerHasExtraTurn = false;
	for (UStatusEffect* StatusEffect : StatusEffects)
	{
		StatusEffect->OnExpiry(PlayerActor);
	}

	HUD->RemoveFromParent();
}

void ATurnBaseCombatV2::BeginPlay()
{
	Super::BeginPlay();

	TArray<AActor*> AllCharacterActorsInScene;

	UGameplayStatics::GetAllActorsOfClassWithTag(GetWorld(), AActor::StaticClass(), FName("Player"), AllCharacterActorsInScene);
	for (AActor* Actor : AllCharacterActorsInScene)
	{
		PlayerActor = Cast<AActor>(Actor);
	}
	PlayerHealth = &Cast<APlayerCharacter>(PlayerActor)->Health;

	HUD = CreateWidget<UUserWidget>(GetWorld(), HUDWidget);

	TurnIndicatorTextBlock = Cast<UTextBlock>(HUD->GetWidgetFromName("TurnIndicator"));
	CurrentComboTextBlock = Cast<UTextBlock>(HUD->GetWidgetFromName("CurrentCombo"));
	ActionPointsTextBlock = Cast<UTextBlock>(HUD->GetWidgetFromName("ActionPoints"));
	BattleLogTextBlock = Cast<UTextBlock>(HUD->GetWidgetFromName("BattleLog"));
	PlayerHealthBar = Cast<UProgressBar>(HUD->GetWidgetFromName("PlayerHealthBar"));
	EnemyHealthBar = Cast<UProgressBar>(HUD->GetWidgetFromName("EnemyHealthBar"));
	ProbertiumResourceBar = Cast<UProgressBar>(HUD->GetWidgetFromName("ProbertiumResourceBar"));
	EisResourceBar = Cast<UProgressBar>(HUD->GetWidgetFromName("EisResourceBar"));
	AzosResourceBar = Cast<UProgressBar>(HUD->GetWidgetFromName("AzosResourceBar"));
	IroquoidResourceBar = Cast<UProgressBar>(HUD->GetWidgetFromName("IroquoidResourceBar"));
	EnemyProbertiumResourceBar = Cast<UProgressBar>(HUD->GetWidgetFromName("ProbertiumResourceBar_E"));
	EnemyEisResourceBar = Cast<UProgressBar>(HUD->GetWidgetFromName("EisResourceBar_E"));
	EnemyAzosResourceBar = Cast<UProgressBar>(HUD->GetWidgetFromName("AzosResourceBar_E"));
	EnemyIroquoidResourceBar = Cast<UProgressBar>(HUD->GetWidgetFromName("IroquoidResourceBar_E"));
	CastButton = Cast<UButton>(HUD->GetWidgetFromName("CastButton"));
	PButton = Cast<UButton>(HUD->GetWidgetFromName("PButton"));
	EButton = Cast<UButton>(HUD->GetWidgetFromName("EButton"));
	AButton = Cast<UButton>(HUD->GetWidgetFromName("AButton"));
	IButton = Cast<UButton>(HUD->GetWidgetFromName("IButton"));
	BackspaceButton = Cast<UButton>(HUD->GetWidgetFromName("BackspaceButton"));
	StealButton = Cast<UButton>(HUD->GetWidgetFromName("StealButton"));
	StealButton->OnClicked.AddDynamic(this, &ATurnBaseCombatV2::StealButtonOnClick);
	CastButton->OnClicked.AddDynamic(this, &ATurnBaseCombatV2::CastButtonOnClick);
	PButton->OnClicked.AddDynamic(this, &ATurnBaseCombatV2::PButtonOnClick);
	EButton->OnClicked.AddDynamic(this, &ATurnBaseCombatV2::EButtonOnClick);
	AButton->OnClicked.AddDynamic(this, &ATurnBaseCombatV2::AButtonOnClick);
	IButton->OnClicked.AddDynamic(this, &ATurnBaseCombatV2::IButtonOnClick);
	BackspaceButton->OnClicked.AddDynamic(this, &ATurnBaseCombatV2::BackspaceButtonOnClick);
	ActionPlanWrapBox = Cast<UWrapBox>(HUD->GetWidgetFromName("ActionPlan_Box"));

	UGameplayStatics::GetAllActorsOfClassWithTag(GetWorld(), AActor::StaticClass(), FName("Enemy"), AllCharacterActorsInScene);
	for (AActor* Actor : AllCharacterActorsInScene)
	{
		StartCombat(Actor);
	}
}

void ATurnBaseCombatV2::ExecuteCast(FString Combo)
{
	if (!IsValidCombo(Combo))
	{
		GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, TEXT("Invalid Combo"));
		for (int i = 0; i < Combo.Len(); i++)
		{
			if (Combo[i] == 'P')
			{
				ProbertiumResource += 1;
			}
			else if (Combo[i] == 'E')
			{
				EisResource += 1;
			}
			else if (Combo[i] == 'A')
			{
				AzosResource += 1;
			}
			else if (Combo[i] == 'I')
			{
				IroquoidResource += 1;
			}
		}
		CurrentComboString = "";
		UpdateComboString(CurrentComboString);
		RevertActionPoints();
		UpdateActionPoints();
		UpdateResourceBars();
		return;
	}
	if (IsSpecialCombo(Combo))
	{
		UStatusSystem* StatusSystem = Cast<UStatusSystem>(PlayerActor->GetComponentByClass(UStatusSystem::StaticClass()));
		if (Combo == "AAE")
		{
			UStatusEffect* TempThornsStatusEffect = NewObject<UStatusEffect>(PlayerActor, ThornsStatusEffect);
			StatusSystem->AddStatusEffect(TempThornsStatusEffect, 1, false);
			StatusEffects.Add(TempThornsStatusEffect);
			AddBattleLogMessage("Player Casted Thorns");
		}
		else if (Combo == "PPI")
		{
			UStatusEffect* TempDOTStatusEffect = NewObject<UStatusEffect>(PlayerActor, DOTStatusEffect);
			StatusSystem->AddStatusEffect(TempDOTStatusEffect, 1, false);
			StatusEffects.Add(TempDOTStatusEffect);
			AddBattleLogMessage("Player Casted DOT");
		}
	}

	CurrentComboString = "";
	UpdateComboString(CurrentComboString);
	RevertActionPoints();
	UpdateActionPoints();

	//Damage Calculation
	switch (bIsPlayerTurn)
	{
	case true:
		// Player Turn
		if (Combo == "")
		{
			DamageEnemy(1, "Punch");
		}
		else
		{
			DamageEnemy(*ValidCombos.Find(Combo), Combo);
		}
		OnPlayerTurn.Broadcast(PlayerActor, EnemyActor);
		break;
	case false:
		break;
	}

	//Ends Combat if either the player or enemy is dead
	if (*EnemyHealth <= 0)
	{
		EndCombat();
		EnemyActor->Destroy();
		CreateWidget<UUserWidget>(GetWorld(), WinWidget)->AddToViewport();
		return;
	}
	if (*PlayerHealth <= 0)
	{
		EndCombat();
		CreateWidget<UUserWidget>(GetWorld(), LoseWidget)->AddToViewport();
		return;
	}

	if (!bPlayerHasExtraTurn)
	{
		SwitchTurn();
	}
	bPlayerHasExtraTurn = false;
}

void ATurnBaseCombatV2::UseActionPoint()
{
	ActiveActionPoints += 1;
	UpdateActionPoints();
}

void ATurnBaseCombatV2::ReuseActionPoint()
{
	ActiveActionPoints -= 1;
	UpdateActionPoints();
}

void ATurnBaseCombatV2::RevertActionPoints()
{
	ActiveActionPoints = 0;
	UpdateActionPoints();
}

void ATurnBaseCombatV2::DamagePlayer(int Damage, FString DamageType)
{
	*PlayerHealth -= FMath::Clamp(Damage, -10, 100);
	UpdateProgressBars();
	AddBattleLogMessage("Player was damaged for " + FString::FromInt(Damage) + " HP by " + DamageType + ".");
}

void ATurnBaseCombatV2::DamageEnemy(int Damage, FString DamageType)
{
	*EnemyHealth -= FMath::Clamp(Damage, -10, 100);
	UpdateProgressBars();
	AddBattleLogMessage("Enemy was damaged for " + FString::FromInt(Damage) + " HP by " + DamageType + ".");
}

void ATurnBaseCombatV2::UpdateProgressBars() const
{
	PlayerHealthBar->SetPercent(*PlayerHealth / 100.0f);
	EnemyHealthBar->SetPercent(*EnemyHealth / 100.0f);
}

bool ATurnBaseCombatV2::IsValidCombo(const FString Combo) const
{
	return ValidCombos.Contains(Combo);
}

bool ATurnBaseCombatV2::IsSpecialCombo(const FString Combo) const
{
	return SpecialCombos.Contains(Combo);
}

void ATurnBaseCombatV2::SwitchTurn()
{
	TurnIndicatorTextBlock->SetText(FText::FromString("Enemy Turn"));
	ToggleButtons();

	FTimerHandle UnusedHandle;
	GetWorldTimerManager().SetTimer(UnusedHandle, this, &ATurnBaseCombatV2::EnemyTurn, 2.0f, false);
}

void ATurnBaseCombatV2::CastButtonOnClick()
{
	ExecuteCast(CurrentComboString);
}

void ATurnBaseCombatV2::PButtonOnClick()
{
	if (ActiveActionPoints >= DefaultActionPoints)
	{
		GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, TEXT("No More Action Points"));
		return;
	}
	if (ProbertiumResource <= 0)
	{
		GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, TEXT("No More Probertium"));
		return;
	}
	UseActionPoint();
	CurrentComboString.AppendChar('P');
	UpdateComboString(CurrentComboString);
	ProbertiumResource -= 1;
	UpdateResourceBars();
}

void ATurnBaseCombatV2::EButtonOnClick()
{
	if (ActiveActionPoints >= DefaultActionPoints)
	{
		GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, TEXT("No More Action Points"));
		return;
	}
	if (EisResource <= 0)
	{
		GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, TEXT("No More Eis"));
		return;
	}
	UseActionPoint();
	CurrentComboString.AppendChar('E');
	UpdateComboString(CurrentComboString);
	EisResource -= 1;
	UpdateResourceBars();
}

void ATurnBaseCombatV2::AButtonOnClick()
{
	if (ActiveActionPoints >= DefaultActionPoints)
	{
		GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, TEXT("No More Action Points"));
		return;
	}
	if (AzosResource <= 0)
	{
		GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, TEXT("No More Azos"));
		return;
	}
	UseActionPoint();
	CurrentComboString.AppendChar('A');
	UpdateComboString(CurrentComboString);
	AzosResource -= 1;
	UpdateResourceBars();
}

void ATurnBaseCombatV2::IButtonOnClick()
{
	if (ActiveActionPoints >= DefaultActionPoints)
	{
		GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, TEXT("No More Action Points"));
		return;
	}
	if (IroquoidResource <= 0)
	{
		GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, TEXT("No More Iroquoid"));
		return;
	}
	UseActionPoint();
	CurrentComboString.AppendChar('I');
	UpdateComboString(CurrentComboString);
	IroquoidResource -= 1;
	UpdateResourceBars();
}

void ATurnBaseCombatV2::BackspaceButtonOnClick()
{
	if (CurrentComboString.Len() <= 0)
	{
		GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, TEXT("Blank Combo"));
		return;
	}
	ReuseActionPoint();
	if (CurrentComboString.Right(1) == "P")
	{
		ProbertiumResource += 1;
	}
	else if (CurrentComboString.Right(1) == "E")
	{
		EisResource += 1;
	}
	else if (CurrentComboString.Right(1) == "A")
	{
		AzosResource += 1;
	}
	else if (CurrentComboString.Right(1) == "I")
	{
		IroquoidResource += 1;
	}
	CurrentComboString.RemoveAt(CurrentComboString.Len() - 1);
	UpdateComboString(CurrentComboString);
	UpdateResourceBars();
}

void ATurnBaseCombatV2::UpdateComboString(FString NewCombo) const
{
	CurrentComboTextBlock->SetText(FText::FromString(NewCombo));
}

void ATurnBaseCombatV2::UpdateActionPoints() const
{
	ActionPointsTextBlock->SetText(FText::FromString(FString::FromInt(DefaultActionPoints - ActiveActionPoints)));
}

void ATurnBaseCombatV2::AddBattleLogMessage(FString Message)
{
	BattleLog.Append(Message + "\n");
	UpdateBattleLog();
}

void ATurnBaseCombatV2::ClearBattleLog()
{
	BattleLog = "";
}

void ATurnBaseCombatV2::UpdateBattleLog()
{
	TArray<FString> TempArray;
	if (const int32 LineCount = BattleLog.ParseIntoArray(TempArray, TEXT("\n"), true); LineCount > 10)
	{
		ClearBattleLog();
	}
	BattleLogTextBlock->SetText(FText::FromString(BattleLog));
}

void ATurnBaseCombatV2::UpdateResourceBars() const
{
	ProbertiumResourceBar->SetPercent(ProbertiumResource / 10.0f);
	EisResourceBar->SetPercent(EisResource / 10.0f);
	AzosResourceBar->SetPercent(AzosResource / 10.0f);
	IroquoidResourceBar->SetPercent(IroquoidResource / 10.0f);
	EnemyProbertiumResourceBar->SetPercent(EnemyProbertiumResource / 10.0f);
	EnemyEisResourceBar->SetPercent(EnemyEisResource / 10.0f);
	EnemyAzosResourceBar->SetPercent(EnemyAzosResource / 10.0f);
	EnemyIroquoidResourceBar->SetPercent(EnemyIroquoidResource / 10.0f);
}

void ATurnBaseCombatV2::ToggleButtons() const
{
	PButton->SetIsEnabled(!PButton->GetIsEnabled());
	EButton->SetIsEnabled(!EButton->GetIsEnabled());
	AButton->SetIsEnabled(!AButton->GetIsEnabled());
	IButton->SetIsEnabled(!IButton->GetIsEnabled());
	BackspaceButton->SetIsEnabled(!BackspaceButton->GetIsEnabled());
	CastButton->SetIsEnabled(!CastButton->GetIsEnabled());
}

void ATurnBaseCombatV2::EnemyTurn()
{
	if (EnemyActionPlan.Num() > 0)
	{
		TMap<FString, int> ActionEffects = EnemyActionPlan[0]->Perform(GetWorldState());
		if (ActionEffects.IsEmpty())
		{
			EnemyActionPlan = EnemyGOAPAgent->Plan(GetWorldState(), EnemyGOAPAgent->Goals);
			ClearActionPlanWidget();
			UpdateActionPlanWidget();
			OnEnemyTurn.Broadcast(EnemyActor, PlayerActor);
			TurnIndicatorTextBlock->SetText(FText::FromString("Player Turn"));
			ToggleButtons();
			return;
		}

		if (ActionEffects.Contains("PlayerHealth"))
		{
			DamagePlayer(ActionEffects["PlayerHealth"], EnemyActionPlan[0]->ActionName);
		}
		else
		{
			DamageEnemy(ActionEffects["EnemyHealth"], EnemyActionPlan[0]->ActionName);
		}

		for (const TTuple<FString, int> Effect : ActionEffects)
		{
			if (Effect.Key == "ProbertiumResource")
			{
				EnemyProbertiumResource -= FMath::RandRange(Effect.Value - 1, Effect.Value + 1);
			}
			else if (Effect.Key == "EisResource")
			{
				EnemyEisResource -= FMath::RandRange(Effect.Value - 1, Effect.Value + 1);
			}
			else if (Effect.Key == "AzosResource")
			{
				EnemyAzosResource -= FMath::RandRange(Effect.Value - 1, Effect.Value + 1);
			}
			else if (Effect.Key == "IroquoidResource")
			{
				EnemyIroquoidResource -= FMath::RandRange(Effect.Value - 1, Effect.Value + 1);
			}
		}

		EnemyActionPlan.RemoveAt(0);
		UpdateResourceBars();
	}
	else
	{
		EnemyActionPlan = EnemyGOAPAgent->Plan(GetWorldState(), EnemyGOAPAgent->Goals);
		ClearActionPlanWidget();
		UpdateActionPlanWidget();
	}
	OnEnemyTurn.Broadcast(EnemyActor, PlayerActor);
	TurnIndicatorTextBlock->SetText(FText::FromString("Player Turn"));
	ToggleButtons();
}

UWorldState* ATurnBaseCombatV2::GetWorldState() const
{
	UWorldState* WorldState = NewObject<UWorldState>();
	WorldState->State.Add("PlayerHealth", *PlayerHealth);
	WorldState->State.Add("EnemyHealth", *EnemyHealth);
	WorldState->State.Add("ProbertiumResource", EnemyProbertiumResource);
	WorldState->State.Add("EisResource", EnemyEisResource);
	WorldState->State.Add("AzosResource", EnemyAzosResource);
	WorldState->State.Add("IroquoidResource", EnemyIroquoidResource);
	return WorldState;
}

void ATurnBaseCombatV2::UpdateActionPlanWidget()
{
	for (UGOAPAction* Action : EnemyActionPlan)
	{
		UUserWidget* ActionWidget = CreateWidget<UUserWidget>(GetWorld(), ActionPlanItemWidget);
		ActionPlanWidgets.Add(ActionWidget);
		Cast<UTextBlock>(ActionWidget->GetWidgetFromName("ActionName"))->SetText(FText::FromString(Action->ActionName));
		ActionPlanWrapBox->AddChild(ActionWidget);
	}
}

void ATurnBaseCombatV2::ClearActionPlanWidget()
{
	for (UUserWidget* ActionWidget : ActionPlanWidgets)
	{
		ActionPlanWrapBox->RemoveChild(ActionWidget);
	}
}

void ATurnBaseCombatV2::StealButtonOnClick()
{
	StealButton->SetIsEnabled(false);
	ClearActionPlanWidget();
	TArray Resources = {&EnemyProbertiumResource, &EnemyEisResource, &EnemyAzosResource, &EnemyIroquoidResource};
	Resources.Sort([](const int& A, const int& B) { return A > B; });
	for (int* Resource : Resources)
	{
		if (*Resource > 0)
		{
			*Resource -= 1;
			if (Resource == &EnemyProbertiumResource)
			{
				ProbertiumResource += 1;
			}
			else if (Resource == &EnemyEisResource)
			{
				EisResource += 1;
			}
			else if (Resource == &EnemyAzosResource)
			{
				AzosResource += 1;
			}
			else if (Resource == &EnemyIroquoidResource)
			{
				IroquoidResource += 1;
			}
			break;
		}
	}
	UpdateResourceBars();
	EnemyActionPlan = EnemyGOAPAgent->Plan(GetWorldState(), EnemyGOAPAgent->Goals);
	UpdateActionPlanWidget();
}
